from typing import Generic, TypeVar

from rainbowadn.chain.block import Block, BlockFactory
from rainbowadn.chain.blockcollectioninterface import BlockCollectionInterface
from rainbowadn.chain.chaincollectioninterface import ChainCollectionInterface
from rainbowadn.chain.reduction.reduction import Reduction, ReductionFactory
from rainbowadn.chain.reduction.reductionprotocol import ReductionProtocol
from rainbowadn.chain.reduction.reductionstageprotocol import ReductionStageProtocol
from rainbowadn.chain.stages.stage import StageStageFactory, StateStage, StateStageFactory
from rainbowadn.chain.stages.stageprotocol import StageProtocol
from rainbowadn.chain.stages.stagestateprotocol import StageStateProtocol
from rainbowadn.chain.states.stateprotocol import StateProtocol
from rainbowadn.hashing.hashpoint import HashPoint
from rainbowadn.hashing.nullability.notnull import NotNull
from rainbowadn.hashing.nullability.null import Null
from rainbowadn.hashing.nullability.nullablereference import NullableReference
from rainbowadn.hashing.rainbow_factory import RainbowFactory

__all__ = ('ReductionChain',)

ReductorType = TypeVar('ReductorType')
AccumulatorType = TypeVar('AccumulatorType')


class ReductionChain(
    ChainCollectionInterface[
        Block[
            ReductorType,
            StateStage[
                ReductorType,
                AccumulatorType,
                Reduction[ReductorType, AccumulatorType]
            ]
        ],
        ReductorType,
        AccumulatorType
    ],
    Generic[ReductorType, AccumulatorType]
):
    def _verify_link(
            self,
            block: Block[
                ReductorType,
                StateStage[
                    ReductorType,
                    AccumulatorType,
                    Reduction[ReductorType, AccumulatorType]
                ]
            ],
            previous: NullableReference[
                Block[
                    ReductorType,
                    StateStage[
                        ReductorType,
                        AccumulatorType,
                        Reduction[ReductorType, AccumulatorType]
                    ]
                ]
            ]
    ) -> bool:
        assert isinstance(block, Block)
        assert isinstance(previous, NullableReference)
        assert block.previous == previous
        previous_chain = ReductionChain(previous, self.reductor_factory, self.accumulator_factory, self.protocol)
        assert isinstance(previous_chain, BlockCollectionInterface)
        state_protocol = StageStateProtocol(self.resolver)
        assert isinstance(state_protocol, StateProtocol)
        return state_protocol.verify(
            previous_chain.state(),
            block.header,
            block.state,
        )

    def _base_state_factory(self) -> RainbowFactory[AccumulatorType]:
        return self.accumulator_factory

    def _base_state(
            self,
            block: Block[
                ReductorType,
                StateStage[
                    ReductorType,
                    AccumulatorType,
                    Reduction[ReductorType, AccumulatorType]
                ]
            ]
    ) -> HashPoint[AccumulatorType]:
        assert isinstance(block, Block)
        return self.resolver.resolve(block.state).state

    def add(
            self,
            header: HashPoint[ReductorType]
    ) -> 'ReductionChain[ReductorType, AccumulatorType]':
        assert isinstance(header, HashPoint)
        return ReductionChain(
            NullableReference.off(
                self._add(header)
            ),
            header.factory,
            self.accumulator_factory,
            self.protocol
        )

    def __init__(
            self,
            reference: NullableReference[
                Block[
                    ReductorType,
                    StateStage[
                        ReductorType,
                        AccumulatorType,
                        Reduction[ReductorType, AccumulatorType]
                    ]
                ]
            ],
            reductor_factory: RainbowFactory[ReductorType],
            accumulator_factory: RainbowFactory[AccumulatorType],
            protocol: ReductionProtocol[ReductorType, AccumulatorType]
    ):
        assert isinstance(reference, NullableReference)
        assert isinstance(reductor_factory, RainbowFactory)
        assert isinstance(accumulator_factory, RainbowFactory)
        assert isinstance(protocol, ReductionProtocol)
        super().__init__(reference, protocol.resolver)
        self.reductor_factory = reductor_factory
        self.accumulator_factory = accumulator_factory
        self.protocol = protocol

    @classmethod
    def _state_stage_factory(
            cls,
            reductor_factory: RainbowFactory[ReductorType],
            accumulator_factory: RainbowFactory[AccumulatorType],
            protocol: ReductionProtocol[ReductorType, AccumulatorType]
    ) -> RainbowFactory[
        StateStage[
            ReductorType,
            AccumulatorType,
            Reduction[ReductorType, AccumulatorType]
        ]
    ]:
        assert isinstance(reductor_factory, RainbowFactory)
        assert isinstance(accumulator_factory, RainbowFactory)
        assert isinstance(protocol, ReductionProtocol)
        stage_protocol = ReductionStageProtocol(protocol)
        assert isinstance(stage_protocol, StageProtocol)
        return StateStageFactory(
            stage_protocol,
            StageStageFactory(
                stage_protocol,
                ReductionFactory(
                    reductor_factory,
                    accumulator_factory
                )
            ),
            accumulator_factory
        )

    @classmethod
    def _block_factory(
            cls,
            reductor_factory: RainbowFactory[ReductorType],
            accumulator_factory: RainbowFactory[AccumulatorType],
            protocol: ReductionProtocol[ReductorType, AccumulatorType]
    ) -> RainbowFactory[
        Block[
            ReductorType,
            StateStage[
                ReductorType,
                AccumulatorType,
                Reduction[ReductorType, AccumulatorType]
            ]
        ]
    ]:
        assert isinstance(reductor_factory, RainbowFactory)
        assert isinstance(accumulator_factory, RainbowFactory)
        assert isinstance(protocol, ReductionProtocol)
        return BlockFactory(
            reductor_factory,
            cls._state_stage_factory(
                reductor_factory,
                accumulator_factory,
                protocol
            )
        )

    @classmethod
    def empty(
            cls,
            reductor_factory: RainbowFactory[ReductorType],
            accumulator_factory: RainbowFactory[AccumulatorType],
            protocol: ReductionProtocol[ReductorType, AccumulatorType]
    ) -> 'ReductionChain[ReductorType, AccumulatorType]':
        assert isinstance(reductor_factory, RainbowFactory)
        assert isinstance(accumulator_factory, RainbowFactory)
        assert isinstance(protocol, ReductionProtocol)
        return cls(
            NullableReference(
                Null(),
                cls._block_factory(
                    reductor_factory,
                    accumulator_factory,
                    protocol
                )
            ),
            reductor_factory,
            accumulator_factory,
            protocol
        )

    def state(
            self
    ) -> NullableReference[
        StateStage[
            ReductorType,
            AccumulatorType,
            Reduction[ReductorType, AccumulatorType]
        ]
    ]:
        if isinstance(self.reference.reference, Null):
            return NullableReference(
                Null(),
                self._state_stage_factory(
                    self.reductor_factory,
                    self.accumulator_factory,
                    self.protocol
                )
            )
        elif isinstance(self.reference.reference, NotNull):
            block: Block[
                ReductorType,
                StateStage[
                    ReductorType,
                    AccumulatorType,
                    Reduction[ReductorType, AccumulatorType]
                ]
            ] = self.resolver.resolve(self.reference.reference.value)
            assert isinstance(block, Block)
            return NullableReference.of(block.state)
        else:
            raise TypeError

    def _add(
            self,
            reductor: HashPoint[ReductorType]
    ) -> Block[
        ReductorType,
        StateStage[
            ReductorType,
            AccumulatorType,
            Reduction[ReductorType, AccumulatorType]
        ]
    ]:
        assert isinstance(reductor, HashPoint)
        stage_protocol = ReductionStageProtocol(self.protocol)
        assert isinstance(stage_protocol, StageProtocol)
        return Block(
            self.reference,
            reductor,
            stage_protocol.derive_full(
                self.state(),
                reductor,
                self.accumulator_factory,
                ReductionFactory(reductor.factory, self.accumulator_factory)
            )
        )

    def previous(
            self
    ):
        assert isinstance(self.reference.reference, NotNull)
        block: Block[
            ReductorType,
            StateStage[
                ReductorType,
                AccumulatorType,
                Reduction[ReductorType, AccumulatorType]
            ]
        ] = self.resolver.resolve(self.reference.reference.value)
        assert isinstance(block, Block)
        return ReductionChain(block.previous, self.reductor_factory, self.accumulator_factory, self.protocol)

    def verify(self) -> bool:
        if isinstance(self.reference.reference, Null):
            return True
        elif isinstance(self.reference.reference, NotNull):
            return self.verify_link(
                self.previous().reference
            ) and self.previous().verify()
        else:
            raise TypeError

    def loose(self) -> ChainCollectionInterface[
        Block[
            ReductorType,
            StateStage[
                ReductorType,
                AccumulatorType,
                Reduction[ReductorType, AccumulatorType]
            ]
        ],
        ReductorType,
        AccumulatorType
    ]:
        return self
