from typing import Generic, Iterable, TypeVar

from rainbowadn.chain.stages.stageprotocol import StageProtocol
from rainbowadn.hashing.hashpoint import HashPoint
from rainbowadn.hashing.hash_point_format import hash_point_format, tabulate
from rainbowadn.hashing.hashresolver import HashResolver
from rainbowadn.hashing.nullability.notnull import NotNull
from rainbowadn.hashing.nullability.null import Null
from rainbowadn.hashing.nullability.nullablereference import NullableReference, NullableReferenceFactory
from rainbowadn.hashing.rainbow_factory import RainbowFactory
from rainbowadn.hashing.recursivementionable import RecursiveMentionable

__all__ = (
    'StageStage',
    'StageStageFactory',
    'StateStage',
    'StateStageFactory',
)

HeaderType = TypeVar('HeaderType')
BaseStateType = TypeVar('BaseStateType')
StageType = TypeVar('StageType')


class Stage(Generic[HeaderType, BaseStateType, StageType]):
    def __init__(self, protocol: StageProtocol[HeaderType, BaseStateType, StageType]):
        assert isinstance(protocol, StageProtocol)
        self.protocol = protocol


class StageStage(
    RecursiveMentionable,
    Stage[HeaderType, BaseStateType, StageType],
    Generic[HeaderType, BaseStateType, StageType],
):
    def __init__(
            self,
            protocol: StageProtocol[HeaderType, BaseStateType, StageType],
            previous: NullableReference['StageStage[HeaderType, BaseStateType, StageType]'],
            stage: HashPoint[StageType],
    ):
        assert isinstance(protocol, StageProtocol)
        assert isinstance(previous, NullableReference)
        assert isinstance(stage, HashPoint)
        super().__init__(protocol)
        self.previous = previous
        self.stage = stage

    def _previous_state(
            self,
            previous: NullableReference['StateStage[HeaderType, BaseStateType, StageType]'],
            base_factory: RainbowFactory[BaseStateType],
    ) -> NullableReference[BaseStateType]:
        assert isinstance(previous, NullableReference)
        assert isinstance(base_factory, RainbowFactory)
        if isinstance(previous.reference, Null):
            return NullableReference(Null(), base_factory)
        elif isinstance(previous.reference, NotNull):
            state_stage: StateStage[HeaderType, BaseStateType, StageType] = self.protocol.resolver.resolve(
                previous.reference.value
            )
            assert isinstance(state_stage, StateStage)
            return NullableReference.of(state_stage.state)
        else:
            raise TypeError

    def verify(
            self,
            previous: NullableReference['StateStage[HeaderType, BaseStateType, StageType]'],
            header: HashPoint[HeaderType],
            base_factory: RainbowFactory[BaseStateType],
    ) -> bool:
        assert isinstance(previous, NullableReference)
        assert isinstance(header, HashPoint)
        assert isinstance(base_factory, RainbowFactory)
        if isinstance(self.previous.reference, Null):
            return self.protocol.verify_header(
                self._previous_state(previous, base_factory),
                header,
                self.stage
            )
        elif isinstance(self.previous.reference, NotNull):
            previous_stage: StageStage[HeaderType, BaseStateType, StageType] = self.protocol.resolver.resolve(
                self.previous.reference.value
            )
            assert isinstance(previous_stage, StageStage)
            return self.protocol.verify_stage(
                previous_stage.stage,
                self.stage
            ) and previous_stage.verify(
                previous,
                header,
                base_factory
            )
        else:
            raise TypeError

    def points(self) -> Iterable[HashPoint]:
        return [*self.previous.points(), self.stage]

    def __bytes__(self):
        return bytes(self.previous) + bytes(self.stage)

    def __factory__(self) -> RainbowFactory['StageStage[HeaderType, BaseStateType, StageType]']:
        return StageStageFactory(
            self.protocol,
            self.stage.factory
        )

    def str(self, resolver: HashResolver, tab: int) -> str:
        assert isinstance(resolver, HashResolver)
        assert isinstance(tab, int)
        return f'{self.previous.str(resolver, tab)}' \
               f'{tabulate(tab)}{hash_point_format(self.stage, resolver, tab)}'


class StageStageFactory(RainbowFactory[StageStage[HeaderType, BaseStateType, StageType]]):
    def __init__(
            self,
            protocol: StageProtocol[HeaderType, BaseStateType, StageType],
            stage_factory: RainbowFactory[StageType]
    ):
        assert isinstance(protocol, StageProtocol)
        assert isinstance(stage_factory, RainbowFactory)
        self.protocol = protocol
        self.stage_factory = stage_factory

    def from_bytes(self, source: bytes) -> StageStage[HeaderType, BaseStateType, StageType]:
        assert isinstance(source, bytes)
        return StageStage(
            self.protocol,
            NullableReferenceFactory(self).from_bytes(source[:HashPoint.HASH_LENGTH]),
            HashPoint(self.stage_factory, source[HashPoint.HASH_LENGTH:], Null())
        )


class StateStage(
    RecursiveMentionable,
    Stage[HeaderType, BaseStateType, StageType],
    Generic[HeaderType, BaseStateType, StageType]
):
    def __init__(
            self,
            protocol: StageProtocol[HeaderType, BaseStateType, StageType],
            previous: HashPoint[StageStage[HeaderType, BaseStateType, StageType]],
            state: HashPoint[BaseStateType]
    ):
        assert isinstance(protocol, StageProtocol)
        assert isinstance(previous, HashPoint)
        assert isinstance(state, HashPoint)
        super().__init__(protocol)
        self.previous = previous
        self.state = state

    def verify(
            self,
            previous: NullableReference['StateStage[HeaderType, BaseStateType, StageType]'],
            header: HashPoint[HeaderType]
    ) -> bool:
        assert isinstance(previous, NullableReference)
        assert isinstance(header, HashPoint)
        previous_stage: StageStage[HeaderType, BaseStateType, StageType] = self.protocol.resolver.resolve(
            self.previous
        )
        assert isinstance(previous_stage, StageStage)
        return self.protocol.verify_state(
            previous_stage.stage,
            self.state
        ) and previous_stage.verify(
            previous,
            header,
            self.state.factory
        )

    def points(self) -> Iterable[HashPoint]:
        return [self.previous, self.state]

    def __bytes__(self):
        return bytes(self.previous) + bytes(self.state)

    def __factory__(self) -> RainbowFactory['StateStage[HeaderType, BaseStateType, StageType]']:
        return StateStageFactory(
            self.protocol,
            self.previous.factory,
            self.state.factory
        )

    def str(self, resolver: HashResolver, tab: int) -> str:
        assert isinstance(resolver, HashResolver)
        assert isinstance(tab, int)
        return f'{hash_point_format(self.previous, resolver, tab)}' \
               f'{tabulate(tab)}{hash_point_format(self.state, resolver, tab)}'


class StateStageFactory(RainbowFactory[StateStage[HeaderType, BaseStateType, StageType]]):
    def __init__(
            self,
            protocol: StageProtocol[HeaderType, BaseStateType, StageType],
            stage_stage_factory: RainbowFactory[StageStage[HeaderType, BaseStateType, StageType]],
            state_factory: RainbowFactory[BaseStateType]
    ):
        assert isinstance(protocol, StageProtocol)
        assert isinstance(stage_stage_factory, RainbowFactory)
        assert isinstance(state_factory, RainbowFactory)
        self.protocol = protocol
        self.stage_stage_factory = stage_stage_factory
        self.state_factory = state_factory

    def from_bytes(self, source: bytes) -> StateStage[HeaderType, BaseStateType, StageType]:
        assert isinstance(source, bytes)
        return StateStage(
            self.protocol,
            HashPoint(self.stage_stage_factory, source[:HashPoint.HASH_LENGTH], Null()),
            HashPoint(self.state_factory, source[HashPoint.HASH_LENGTH:], Null()),
        )
