from typing import Generic, Iterable, TypeVar

from rainbowadn.hashing.hash_point_format import hash_point_format, tabulate
from rainbowadn.hashing.hashpoint import HashPoint
from rainbowadn.hashing.hashresolver import HashResolver
from rainbowadn.hashing.nullability.null import Null
from rainbowadn.hashing.nullability.nullablereference import NullableReference, NullableReferenceFactory
from rainbowadn.hashing.rainbow_factory import RainbowFactory
from rainbowadn.hashing.recursivementionable import RecursiveMentionable

__all__ = ('Stack', 'StackFactory',)

ElementType = TypeVar('ElementType')


class Stack(RecursiveMentionable, Generic[ElementType]):
    def __factory__(self) -> RainbowFactory['Stack[ElementType]']:
        return self.factory(self.element.factory)

    @classmethod
    def factory(cls, factory: RainbowFactory[ElementType]) -> RainbowFactory['Stack[ElementType]']:
        assert isinstance(factory, RainbowFactory)
        return StackFactory(factory)

    def __init__(self, previous: NullableReference['Stack[ElementType]'], element: HashPoint[ElementType]):
        assert isinstance(previous, NullableReference)
        assert isinstance(element, HashPoint)
        self.previous = previous
        self.element = element

    def points(self) -> Iterable[HashPoint]:
        return [*self.previous.points(), self.element]

    def __bytes__(self):
        return bytes(self.previous) + bytes(self.element)

    def str(self, resolver: HashResolver, tab: int) -> str:
        assert isinstance(resolver, HashResolver)
        assert isinstance(tab, int)
        return f'{self.previous.str(resolver, tab)}' \
               f'{tabulate(tab)}{hash_point_format(self.element, resolver, tab)}'

    @classmethod
    def of(
            cls,
            factory: RainbowFactory[ElementType],
            elements: Iterable[HashPoint[ElementType]]
    ) -> NullableReference[
        'Stack[ElementType]'
    ]:
        assert isinstance(factory, RainbowFactory)
        reference: NullableReference[Stack[ElementType]] = NullableReference(Null(), cls.factory(factory))
        for element in elements:
            assert isinstance(element, HashPoint)
            reference = NullableReference.off(cls(reference, element))
        return reference

    @classmethod
    def off(
            cls,
            factory: RainbowFactory[ElementType],
            elements: Iterable[ElementType]
    ) -> NullableReference[
        'Stack[ElementType]'
    ]:
        return cls.of(factory, map(HashPoint.of, elements))


class StackFactory(RainbowFactory[Stack[ElementType]], Generic[ElementType]):
    def __init__(self, factory: RainbowFactory[ElementType]):
        assert isinstance(factory, RainbowFactory)
        self.factory = factory

    def from_bytes(self, source: bytes) -> Stack[ElementType]:
        assert isinstance(source, bytes)
        return Stack(
            NullableReferenceFactory(self).from_bytes(source[:HashPoint.HASH_LENGTH]),
            HashPoint(self.factory, source[HashPoint.HASH_LENGTH:], Null())
        )

    def loose(self) -> RainbowFactory[Stack[ElementType]]:
        return self
