from typing import Generic, TypeVar

from rainbowadn.data.atomic.integer import Integer
from rainbowadn.data.collection.trees.binary.activebinarytree import ActiveBinaryTree
from rainbowadn.data.collection.trees.binary.binarytree import BinaryTree, BinaryTreeFactory
from rainbowadn.data.collection.trees.comparison.comparator import Comparator
from rainbowadn.data.collection.keymetadata import KeyMetadata, KeyMetadataFactory
from rainbowadn.hashing.hashpoint import HashPoint
from rainbowadn.hashing.nullability.notnull import NotNull
from rainbowadn.hashing.nullability.null import Null
from rainbowadn.hashing.nullability.nullable import Nullable
from rainbowadn.hashing.nullability.nullablereference import NullableReference
from rainbowadn.hashing.rainbow_factory import RainbowFactory

__all__ = ('AVLABT',)

ActiveKeyType = TypeVar('ActiveKeyType')


class AVLABT(ActiveBinaryTree[ActiveKeyType, Integer], Generic[ActiveKeyType]):
    def __init__(
            self,
            comparator: Comparator[ActiveKeyType],
            reference: NullableReference[BinaryTree[KeyMetadata[ActiveKeyType, Integer]]]
    ):
        assert isinstance(comparator, Comparator)
        assert isinstance(reference, NullableReference)
        super().__init__(comparator, reference)

    def create(
            self, reference: NullableReference[BinaryTree[KeyMetadata[ActiveKeyType, Integer]]]
    ) -> 'AVLABT[ActiveKeyType]':
        assert isinstance(reference, NullableReference)
        return type(self)(
            self.comparator,
            reference
        )

    def height(self) -> int:
        reference: Nullable[
            HashPoint[BinaryTree[KeyMetadata[ActiveKeyType, Integer]]]
        ] = self.reference.reference
        assert isinstance(reference, Nullable)
        if isinstance(reference, Null):
            return 0
        elif isinstance(reference, NotNull):
            tree: BinaryTree[KeyMetadata[ActiveKeyType, Integer]] = self.resolver.resolve(reference.value)
            assert isinstance(tree, BinaryTree)
            key: KeyMetadata[ActiveKeyType, Integer] = self.resolver.resolve(tree.key)
            assert isinstance(key, KeyMetadata)
            return self.resolver.resolve(key.metadata).integer

    @classmethod
    def empty(
            cls,
            comparator: Comparator[ActiveKeyType],
            factory: RainbowFactory[ActiveKeyType]
    ):
        assert isinstance(comparator, Comparator)
        assert isinstance(factory, RainbowFactory)
        return cls(
            comparator,
            NullableReference(
                Null(),
                BinaryTreeFactory(KeyMetadataFactory(factory, Integer.factory()))
            )
        )

    def empty_metadata(self) -> HashPoint[Integer]:
        return HashPoint.of(Integer(0))

    def metadata(
            self,
            treel: NullableReference[BinaryTree[KeyMetadata[ActiveKeyType, Integer]]],
            treer: NullableReference[BinaryTree[KeyMetadata[ActiveKeyType, Integer]]],
            key: HashPoint[ActiveKeyType]
    ) -> HashPoint[Integer]:
        assert isinstance(treel, NullableReference)
        assert isinstance(treer, NullableReference)
        assert isinstance(key, HashPoint)
        return HashPoint.of(Integer(1 + max(self.create(treel).height(), self.create(treer).height())))

    def balance(self) -> ActiveBinaryTree[ActiveKeyType, Integer]:
        reference: Nullable[
            HashPoint[BinaryTree[KeyMetadata[ActiveKeyType, Integer]]]
        ] = self.reference.reference
        assert isinstance(reference, Nullable)
        if isinstance(reference, Null):
            return self
        elif isinstance(reference, NotNull):
            tree: BinaryTree[KeyMetadata[ActiveKeyType, Integer]] = self.resolver.resolve(reference.value)
            assert isinstance(tree, BinaryTree)
            atl = self.create(tree.treel)
            assert isinstance(atl, ActiveBinaryTree)
            atr = self.create(tree.treer)
            assert isinstance(atr, ActiveBinaryTree)
            delta = atl.height() - atr.height()
            assert isinstance(delta, int)
            if delta < -1:
                assert isinstance(atr.reference.reference, NotNull)
                treer: BinaryTree[
                    KeyMetadata[ActiveKeyType, Integer]
                ] = self.resolver.resolve(atr.reference.reference.value)
                assert isinstance(treer, BinaryTree)
                atrl = self.create(treer.treel)
                assert isinstance(atrl, ActiveBinaryTree)
                atrr = self.create(treer.treer)
                assert isinstance(atrr, ActiveBinaryTree)
                if atrl.height() > atrr.height():
                    assert isinstance(atrl.reference.reference, NotNull)
                    treerl: BinaryTree[
                        KeyMetadata[ActiveKeyType, Integer]
                    ] = self.resolver.resolve(atrl.reference.reference.value)
                    assert isinstance(treerl, BinaryTree)
                    return self.active_tree(
                        self.active_tree(
                            atl.reference,
                            treerl.treel,
                            self.resolver.resolve(tree.key).key
                        ).reference,
                        self.active_tree(
                            treerl.treer,
                            atrr.reference,
                            self.resolver.resolve(treer.key).key
                        ).reference,
                        self.resolver.resolve(treerl.key).key
                    )
                else:
                    return self.active_tree(
                        self.active_tree(
                            atl.reference,
                            atrl.reference,
                            self.resolver.resolve(tree.key).key
                        ).reference,
                        atrr.reference,
                        self.resolver.resolve(treer.key).key
                    )
            elif delta > 1:
                assert isinstance(atl.reference.reference, NotNull)
                treel: BinaryTree[
                    KeyMetadata[ActiveKeyType, Integer]
                ] = self.resolver.resolve(atl.reference.reference.value)
                assert isinstance(treel, BinaryTree)
                atlr = self.create(treel.treer)
                assert isinstance(atlr, ActiveBinaryTree)
                atll = self.create(treel.treel)
                assert isinstance(atll, ActiveBinaryTree)
                if atlr.height() > atll.height():
                    assert isinstance(atlr.reference.reference, NotNull)
                    treelr: BinaryTree[
                        KeyMetadata[ActiveKeyType, Integer]
                    ] = self.resolver.resolve(atlr.reference.reference.value)
                    assert isinstance(treelr, BinaryTree)
                    return self.active_tree(
                        self.active_tree(
                            atll.reference,
                            treelr.treel,
                            self.resolver.resolve(treel.key).key
                        ).reference,
                        self.active_tree(
                            treelr.treer,
                            atr.reference,
                            self.resolver.resolve(tree.key).key
                        ).reference,
                        self.resolver.resolve(treelr.key).key
                    )
                else:
                    return self.active_tree(
                        atll.reference,
                        self.active_tree(
                            atlr.reference,
                            atr.reference,
                            self.resolver.resolve(tree.key).key
                        ).reference,
                        self.resolver.resolve(treel.key).key
                    )
            else:
                return self
        else:
            raise TypeError
