from typing import Generic, Iterable, TypeVar

from rainbowadn.hashing.hashpoint import HashPoint
from rainbowadn.hashing.hash_point_format import hash_point_format, tabulate
from rainbowadn.hashing.hashresolver import HashResolver
from rainbowadn.hashing.nullability.null import Null
from rainbowadn.hashing.nullability.nullablereference import NullableReference, NullableReferenceFactory
from rainbowadn.hashing.rainbow_factory import RainbowFactory
from rainbowadn.hashing.recursivementionable import RecursiveMentionable

__all__ = ('BinaryTree', 'BinaryTreeFactory',)

TreeKeyType = TypeVar('TreeKeyType')


class BinaryTree(RecursiveMentionable, Generic[TreeKeyType]):
    def __factory__(self) -> RainbowFactory['BinaryTree[TreeKeyType]']:
        return self.factory(self.key.factory)

    @classmethod
    def factory(cls, factory: RainbowFactory[TreeKeyType]) -> RainbowFactory['BinaryTree[TreeKeyType]']:
        assert isinstance(factory, RainbowFactory)
        return BinaryTreeFactory(factory)

    def __init__(
            self,
            treel: NullableReference['BinaryTree[TreeKeyType]'], treer: NullableReference['BinaryTree[TreeKeyType]'],
            key: HashPoint[TreeKeyType]
    ):
        assert isinstance(treel, NullableReference)
        assert isinstance(treer, NullableReference)
        assert isinstance(key, HashPoint)
        self.treel = treel
        self.treer = treer
        self.key = key

    def points(self) -> Iterable[HashPoint]:
        return [*self.treel.points(), *self.treer.points(), self.key]

    def __bytes__(self):
        return bytes(self.treel) + bytes(self.treer) + bytes(self.key)

    def str(self, resolver: HashResolver, tab: int) -> str:
        assert isinstance(resolver, HashResolver)
        assert isinstance(tab, int)
        return f'{self.treel.str(resolver, tab)}' \
               f'{tabulate(tab)}{hash_point_format(self.key, resolver, tab)}' \
               f'{tabulate(tab)}{self.treer.str(resolver, tab)}'


class BinaryTreeFactory(RainbowFactory[BinaryTree[TreeKeyType]], Generic[TreeKeyType]):
    def __init__(self, factory: RainbowFactory[TreeKeyType]):
        assert isinstance(factory, RainbowFactory)
        self.factory = factory

    def from_bytes(self, source: bytes) -> BinaryTree[TreeKeyType]:
        assert isinstance(source, bytes)
        return BinaryTree(
            NullableReferenceFactory(self).from_bytes(source[:HashPoint.HASH_LENGTH]),
            NullableReferenceFactory(self).from_bytes(source[HashPoint.HASH_LENGTH:HashPoint.HASH_LENGTH * 2]),
            HashPoint(self.factory, source[HashPoint.HASH_LENGTH * 2:], Null())
        )

    def loose(self) -> RainbowFactory[BinaryTree[TreeKeyType]]:
        return self
