from typing import Iterable

from rainbowadn.data.atomic.integer import Integer
from rainbowadn.data.collection.keymetadata import KeyMetadata, KeyMetadataFactory
from rainbowadn.data.collection.trees.binary.activebinarytree import ActiveBinaryTree
from rainbowadn.data.collection.trees.binary.avl import AVLABT
from rainbowadn.data.collection.trees.binary.binarytree import BinaryTree, BinaryTreeFactory
from rainbowadn.data.collection.trees.comparison.comparator import Fail
from rainbowadn.data.collection.trees.comparison.hashcomparator import HashComparator
from rainbowadn.hashing.hash_point_format import tabulate
from rainbowadn.hashing.hashpoint import HashPoint
from rainbowadn.hashing.hashresolver import HashResolver
from rainbowadn.hashing.nullability.notnull import NotNull
from rainbowadn.hashing.nullability.null import Null
from rainbowadn.hashing.nullability.nullable import Nullable
from rainbowadn.hashing.nullability.nullablereference import NullableReference, NullableReferenceFactory
from rainbowadn.hashing.rainbow_factory import RainbowFactory
from rainbowadn.hashing.recursivementionable import RecursiveMentionable
from rainbowadn.hashing.static import StaticFactory, StaticMentionable
from rainbowadn.v13.algo import MINT_CONST
from rainbowadn.v13.subject import Subject
from rainbowadn.v13.transaction import Coin, Transaction, TransactionData

__all__ = ('BankState',)


class BankState(RecursiveMentionable, StaticMentionable['BankState']):
    def __init__(
            self,
            minted: NullableReference[BinaryTree[KeyMetadata[Coin, Integer]]],
            used: NullableReference[BinaryTree[KeyMetadata[Coin, Integer]]],
            miner: NullableReference[Subject],
            length: HashPoint[Integer]
    ):
        assert isinstance(miner, NullableReference)
        assert isinstance(used, NullableReference)
        assert isinstance(miner, NullableReference)
        assert isinstance(length, HashPoint)
        self.minted = minted
        self.used = used
        self.miner = miner
        self.length = length

    def points(self) -> Iterable[HashPoint]:
        return [*self.minted.points(), *self.used.points(), *self.miner.points(), self.length]

    def __bytes__(self):
        return bytes(self.minted) + bytes(self.used) + bytes(self.miner) + bytes(self.length)

    @classmethod
    def from_bytes(cls, source: bytes) -> 'BankState':
        assert isinstance(source, bytes)
        reference_factory: RainbowFactory[
            NullableReference[BinaryTree[KeyMetadata[Coin, Integer]]]
        ] = NullableReferenceFactory(
            BinaryTreeFactory(
                KeyMetadataFactory(
                    StaticFactory(Coin),
                    StaticFactory(Integer),
                ).loose()
            ).loose()
        ).loose()
        assert isinstance(reference_factory, RainbowFactory)
        return cls(
            reference_factory.from_bytes(source[:HashPoint.HASH_LENGTH]),
            reference_factory.from_bytes(source[HashPoint.HASH_LENGTH:2 * HashPoint.HASH_LENGTH]),
            NullableReferenceFactory(
                StaticFactory(Subject)
            ).from_bytes(source[2 * HashPoint.HASH_LENGTH:3 * HashPoint.HASH_LENGTH]),
            HashPoint(Integer.factory(), source[3 * HashPoint.HASH_LENGTH:], Null())
        )

    def loose(self) -> 'BankState':
        return BankState(
            self.minted,
            self.used,
            NullableReference(Null(), self.miner.factory),
            self.length
        )

    def advance(self, resolver: HashResolver) -> 'BankState':
        assert isinstance(resolver, HashResolver)
        return BankState(
            self.minted,
            self.used,
            NullableReference(Null(), self.miner.factory),
            HashPoint.of(Integer(resolver.resolve(self.length).integer + 1))
        )

    def push(self, transaction: HashPoint[Transaction], resolver: HashResolver) -> 'BankState':
        assert isinstance(transaction, HashPoint)
        assert isinstance(resolver, HashResolver)

        transaction_resolved = resolver.resolve(transaction)
        assert isinstance(transaction_resolved, Transaction)

        miner: Nullable[HashPoint[Subject]] = self.miner.reference
        assert isinstance(miner, Nullable)
        if isinstance(miner, Null):
            mint = MINT_CONST
        elif isinstance(miner, NotNull):
            mint = 0
        else:
            raise TypeError

        assert transaction_resolved.verify(resolver, mint)
        transaction_data: TransactionData = resolver.resolve(transaction_resolved.data)
        assert isinstance(transaction_data, TransactionData)

        minted: ActiveBinaryTree[Coin, Integer] = AVLABT(HashComparator(resolver, Fail()), self.minted)
        assert isinstance(minted, ActiveBinaryTree)
        used: ActiveBinaryTree[Coin, Integer] = AVLABT(HashComparator(resolver, Fail()), self.used)
        assert isinstance(used, ActiveBinaryTree)

        in_coin: HashPoint[Coin]
        for in_coin in transaction_data.iter_in_coins(resolver):
            assert isinstance(in_coin, HashPoint)
            assert isinstance(minted.query_tree().query(in_coin), NotNull)
            assert isinstance(used.query_tree().query(in_coin), Null)
            used = used.add(in_coin)
            assert isinstance(used, ActiveBinaryTree)

        coin: Coin
        for coin, miner in transaction_resolved.iter_coins(resolver, mint, miner):
            assert isinstance(coin, Coin)
            assert isinstance(miner, Nullable)
            assert isinstance(minted.query_tree().query(HashPoint.of(coin)), Null)
            minted = minted.add(HashPoint.of(coin))
            assert isinstance(minted, ActiveBinaryTree)

        return BankState(
            minted.reference,
            used.reference,
            NullableReference(miner, self.miner.factory),
            self.length
        )

    def str(self, resolver: HashResolver, tab: int) -> str:
        assert isinstance(resolver, HashResolver)
        assert isinstance(tab, int)
        return f'(' \
               f'{tabulate(tab + 1)}bank' \
               f'{tabulate(tab + 1)}(miner)' \
               f'{tabulate(tab + 1)}{self.miner.str(resolver, tab + 1)}' \
               f'{tabulate(tab + 1)}(minted)' \
               f'{tabulate(tab + 1)}{self.minted.str(resolver, tab + 1)}' \
               f'{tabulate(tab + 1)}(used)' \
               f'{tabulate(tab + 1)}{self.used.str(resolver, tab + 1)}' \
               f'{tabulate(tab + 1)}(length)' \
               f'{tabulate(tab + 1)}{resolver.resolve(self.length)}' \
               f'{tabulate(tab)})'
