from typing import Iterable

import nacl.signing

from rainbowadn.data.atomic.integer import Integer
from rainbowadn.data.collection.stack.stack import Stack, StackFactory
from rainbowadn.hashing.hash_point_format import hash_point_format, tabulate
from rainbowadn.hashing.hashpoint import HashPoint
from rainbowadn.hashing.hashresolver import HashResolver
from rainbowadn.hashing.nullability.notnull import NotNull
from rainbowadn.hashing.nullability.null import Null
from rainbowadn.hashing.nullability.nullable import Nullable
from rainbowadn.hashing.nullability.nullablereference import NullableReference, NullableReferenceFactory
from rainbowadn.hashing.rainbow_factory import RainbowFactory
from rainbowadn.hashing.recursivementionable import RecursiveMentionable
from rainbowadn.hashing.static import StaticMentionable
from rainbowadn.v13.signature import Signature
from rainbowadn.v13.subject import Subject

__all__ = ('CoinData', 'Coin', 'TransactionData', 'Transaction',)


class CoinData(RecursiveMentionable, StaticMentionable['CoinData']):
    def __init__(
            self,
            owner: HashPoint[Subject],
            value: HashPoint[Integer]
    ):
        assert isinstance(owner, HashPoint)
        assert isinstance(value, HashPoint)
        self.owner = owner
        self.value = value

    @classmethod
    def of(cls, owner: Subject, value: int) -> 'CoinData':
        return cls(HashPoint.of(owner), HashPoint.of(Integer(value)))

    def points(self) -> Iterable[HashPoint]:
        return [self.owner, self.value]

    def __bytes__(self):
        return bytes(self.owner) + bytes(self.value)

    @classmethod
    def from_bytes(cls, source: bytes) -> 'CoinData':
        assert isinstance(source, bytes)
        return cls(
            HashPoint(Subject.factory(), source[:HashPoint.HASH_LENGTH], Null()),
            HashPoint(Integer.factory(), source[HashPoint.HASH_LENGTH:], Null()),
        )

    def str(self, resolver: HashResolver, tab: int) -> str:
        assert isinstance(resolver, HashResolver)
        assert isinstance(tab, int)
        return f'{resolver.resolve(self.owner)}' \
               f'{tabulate(tab)}{resolver.resolve(self.value)}'


class Coin(RecursiveMentionable, StaticMentionable['Coin']):
    def __init__(
            self,
            data: HashPoint[CoinData],
            origin: HashPoint['Transaction'],
            index: HashPoint[Integer]
    ):
        assert isinstance(data, HashPoint)
        assert isinstance(origin, HashPoint)
        assert isinstance(index, HashPoint)
        self.data = data
        self.origin = origin
        self.index = index

    def points(self) -> Iterable[HashPoint]:
        return [self.data, self.origin, self.index]

    def __bytes__(self):
        return bytes(self.data) + bytes(self.origin) + bytes(self.index)

    @classmethod
    def from_bytes(cls, source: bytes) -> 'Coin':
        assert isinstance(source, bytes)
        return cls(
            HashPoint(CoinData.factory(), source[:HashPoint.HASH_LENGTH], Null()),
            HashPoint(Transaction.factory(), source[HashPoint.HASH_LENGTH:2 * HashPoint.HASH_LENGTH], Null()),
            HashPoint(Integer.factory(), source[2 * HashPoint.HASH_LENGTH:], Null()),
        )

    def __str__(self):
        return f'(coin)'

    def str(self, resolver: HashResolver, tab: int) -> str:
        assert isinstance(resolver, HashResolver)
        assert isinstance(tab, int)
        return f'(' \
               f'{tabulate(tab + 1)}coin' \
               f'{tabulate(tab + 1)}{hash_point_format(self.data, resolver, tab + 1)}' \
               f'{tabulate(tab + 1)}(origin)' \
               f'{tabulate(tab + 1)}{resolver.resolve(self.index)}' \
               f'{tabulate(tab)})'


class TransactionData(RecursiveMentionable, StaticMentionable['TransactionData']):
    def __init__(
            self,
            in_coins: NullableReference[Stack[Coin]],
            out_coins: NullableReference[Stack[CoinData]],
    ):
        assert isinstance(in_coins, NullableReference)
        assert isinstance(out_coins, NullableReference)
        self.in_coins = in_coins
        self.out_coins = out_coins
        self.hash_point = HashPoint.of(self)
        assert isinstance(self.hash_point, HashPoint)

    def points(self) -> Iterable[HashPoint]:
        return [*self.in_coins.points(), *self.out_coins.points()]

    def __bytes__(self):
        return bytes(self.in_coins) + bytes(self.out_coins)

    @classmethod
    def from_bytes(cls, source: bytes) -> 'TransactionData':
        assert isinstance(source, bytes)
        return cls(
            NullableReferenceFactory(
                StackFactory(Coin.factory()).loose()
            ).from_bytes(source[:HashPoint.HASH_LENGTH]),
            NullableReferenceFactory(
                StackFactory(CoinData.factory()).loose()
            ).from_bytes(source[HashPoint.HASH_LENGTH:]),
        )

    def _verify_signatures(
            self,
            resolver: HashResolver,
            in_coins: NullableReference[Stack[Coin]],
            signatures: NullableReference[Stack[Signature]],
    ) -> bool:
        assert isinstance(resolver, HashResolver)
        assert isinstance(in_coins, NullableReference)
        assert isinstance(signatures, NullableReference)
        if isinstance(in_coins.reference, Null):
            assert isinstance(signatures.reference, Null)
            return True
        elif isinstance(in_coins.reference, NotNull):
            assert isinstance(signatures.reference, NotNull)
            in_coins_stack: Stack[Coin] = resolver.resolve(in_coins.reference.value)
            assert isinstance(in_coins_stack, Stack)
            signatures_stack: Stack[Signature] = resolver.resolve(signatures.reference.value)
            assert isinstance(signatures_stack, Stack)
            coin: Coin = resolver.resolve(in_coins_stack.element)
            assert isinstance(coin, Coin)
            coin_data: CoinData = resolver.resolve(coin.data)
            assert isinstance(coin_data, CoinData)
            owner: Subject = resolver.resolve(coin_data.owner)
            assert isinstance(owner, Subject)
            signature: Signature = resolver.resolve(signatures_stack.element)
            assert isinstance(signature, Signature)
            return signature.verify(
                owner,
                self.hash_point
            )
        else:
            raise TypeError

    def iter_in_coins(self, resolver: HashResolver) -> Iterable[HashPoint[Coin]]:
        assert isinstance(resolver, HashResolver)
        in_coins: NullableReference[Stack[Coin]] = self.in_coins
        assert isinstance(in_coins, NullableReference)
        while isinstance(in_coins.reference, NotNull):
            in_coins_stack: Stack[Coin] = resolver.resolve(in_coins.reference.value)
            assert isinstance(in_coins_stack, Stack)
            coin: HashPoint[Coin] = in_coins_stack.element
            assert isinstance(coin, HashPoint)
            yield coin
            in_coins = in_coins_stack.previous
            assert isinstance(in_coins, NullableReference)
        assert isinstance(in_coins.reference, Null)

    def _total_in(self, resolver: HashResolver) -> int:
        assert isinstance(resolver, HashResolver)
        total_in = 0
        coin: HashPoint[Coin]
        for coin in self.iter_in_coins(resolver):
            assert isinstance(coin, HashPoint)
            total_in += resolver.resolve(
                resolver.resolve(
                    resolver.resolve(
                        coin
                    ).data
                ).value
            ).integer
        return total_in

    def iter_out_coins(self, resolver: HashResolver) -> Iterable[HashPoint[CoinData]]:
        assert isinstance(resolver, HashResolver)
        out_coins: NullableReference[Stack[CoinData]] = self.out_coins
        assert isinstance(out_coins, NullableReference)
        while isinstance(out_coins.reference, NotNull):
            out_coins_stack: Stack[CoinData] = resolver.resolve(out_coins.reference.value)
            assert isinstance(out_coins_stack, Stack)
            coin: HashPoint[CoinData] = out_coins_stack.element
            assert isinstance(coin, HashPoint)
            yield coin
            out_coins = out_coins_stack.previous
            assert isinstance(out_coins, NullableReference)
        assert isinstance(out_coins.reference, Null)

    def _total_out(self, resolver: HashResolver) -> int:
        assert isinstance(resolver, HashResolver)
        total_out = 0
        coin: HashPoint[CoinData]
        for coin in self.iter_out_coins(resolver):
            assert isinstance(coin, HashPoint)
            total_out += resolver.resolve(
                resolver.resolve(
                    coin
                ).value
            ).integer
        return total_out

    def _verify_values(self, resolver: HashResolver, mint: int) -> bool:
        assert isinstance(resolver, HashResolver)
        assert isinstance(mint, int)
        assert self._total_out(resolver) <= self._total_in(resolver) + mint
        return True

    def extra(self, resolver: HashResolver, mint: int) -> int:
        assert isinstance(resolver, HashResolver)
        assert isinstance(mint, int)
        return self._total_in(resolver) + mint - self._total_out(resolver)

    def verify(
            self,
            resolver: HashResolver,
            signatures: NullableReference[Stack[Signature]],
            mint: int
    ) -> bool:
        assert isinstance(resolver, HashResolver)
        assert isinstance(signatures, NullableReference)
        assert isinstance(mint, int)
        return self._verify_signatures(resolver, self.in_coins, signatures) and self._verify_values(resolver, mint)

    def str(self, resolver: HashResolver, tab: int) -> str:
        assert isinstance(resolver, HashResolver)
        assert isinstance(tab, int)
        return f'(in)' \
               f'{tabulate(tab)}{self.in_coins.str(resolver, tab)}' \
               f'{tabulate(tab)}(out)' \
               f'{tabulate(tab)}{self.out_coins.str(resolver, tab)}'


class Transaction(RecursiveMentionable, StaticMentionable['Transaction']):
    def __init__(
            self,
            data: HashPoint[TransactionData],
            signatures: NullableReference[Stack[Signature]]
    ):
        assert isinstance(data, HashPoint)
        assert isinstance(signatures, NullableReference)
        self.data = data
        self.signatures = signatures
        self.hash_point = HashPoint.of(self)
        assert isinstance(self.hash_point, HashPoint)

    def points(self) -> Iterable[HashPoint]:
        return [self.data, *self.signatures.points()]

    def __bytes__(self):
        return bytes(self.data) + bytes(self.signatures)

    @classmethod
    def from_bytes(cls, source: bytes) -> 'Transaction':
        assert isinstance(source, bytes)
        signature_factory: RainbowFactory[Signature] = Signature.factory()
        assert isinstance(signature_factory, RainbowFactory)
        stack_factory: RainbowFactory[Stack[Signature]] = StackFactory(signature_factory).loose()
        assert isinstance(stack_factory, RainbowFactory)
        return cls(
            HashPoint(TransactionData.factory(), source[:HashPoint.HASH_LENGTH], Null()),
            NullableReferenceFactory(stack_factory).from_bytes(source[HashPoint.HASH_LENGTH:]),
        )

    def iter_coins(
            self,
            resolver: HashResolver,
            mint: int,
            miner: Nullable[HashPoint[Subject]]
    ) -> Iterable[tuple[Coin, Nullable[HashPoint[Subject]]]]:
        transaction_data = resolver.resolve(self.data)
        assert isinstance(transaction_data, TransactionData)
        index = 0
        out_coin: HashPoint[CoinData]
        for out_coin in transaction_data.iter_out_coins(resolver):
            assert isinstance(out_coin, HashPoint)
            if isinstance(miner, Null):
                miner = NotNull(
                    resolver.resolve(out_coin).owner
                )
                assert isinstance(miner, Nullable)
            coin: Coin = Coin(out_coin, self.hash_point, HashPoint.of(Integer(index)))
            assert isinstance(coin, Coin)
            yield coin, miner
            index += 1
        if isinstance(miner, NotNull):
            coin: Coin = Coin(
                HashPoint.of(
                    CoinData(
                        miner.value,
                        HashPoint.of(Integer(transaction_data.extra(resolver, mint)))
                    )
                ),
                self.hash_point,
                HashPoint.of(Integer(index))
            )
            assert isinstance(coin, Coin)
            yield coin, miner

    def coins(
            self,
            resolver: HashResolver,
            mint: int,
            miner: Nullable[HashPoint[Subject]]
    ) -> list[Coin]:
        return [coin for coin, _ in self.iter_coins(resolver, mint, miner)]

    def verify(self, resolver: HashResolver, mint: int):
        assert isinstance(resolver, HashResolver)
        assert isinstance(mint, int)
        data: TransactionData = resolver.resolve(self.data)
        assert isinstance(data, TransactionData)
        return data.verify(resolver, self.signatures, mint)

    def __str__(self):
        return f'(transaction)'

    def str(self, resolver: HashResolver, tab: int) -> str:
        assert isinstance(resolver, HashResolver)
        assert isinstance(tab, int)
        return f'(' \
               f'{tabulate(tab + 1)}transaction' \
               f'{tabulate(tab + 1)}{hash_point_format(self.data, resolver, tab + 1)}' \
               f'{tabulate(tab + 1)}{self.signatures.str(resolver, tab + 1)}' \
               f'{tabulate(tab)})'

    @classmethod
    def make(
            cls,
            in_coins: list[Coin],
            out_coins: list[CoinData],
            keys: list[nacl.signing.SigningKey],
    ) -> 'Transaction':
        transaction_data = TransactionData(
            Stack.off(Coin.factory(), reversed(in_coins)),
            Stack.off(CoinData.factory(), reversed(out_coins)),
        )
        assert isinstance(transaction_data, TransactionData)
        return Transaction(
            HashPoint.of(transaction_data),
            Stack.off(
                Signature.factory(),
                (Signature.sign(key, HashPoint.of(transaction_data)) for key in reversed(keys))
            )
        )
